import std.stdio;
import std.thread;
import std.gc;
import win32.windows;
import win32frame.MainWindow ;
import win32frame.glut;

import GameInfo;
import Sprite;

//typedef int function(GLint) pfwglSwapIntervalEXT;
//pfwglSwapIntervalEXT wglSwapIntervalEXT;


ushort LOWORD(uint l){ return l & 0xffff; }
ushort HIWORD(uint l){ return l >>> 16; }

class GameFrame : MainWindow
{
  HINSTANCE hInstance ;
  HGLRC hRC;
  HDC   WinDC;
  PIXELFORMATDESCRIPTOR pfd = { 
    PIXELFORMATDESCRIPTOR.sizeof,    // size of this pfd 
    1,                               // version number 
    PFD_DRAW_TO_WINDOW |             // support window 
    PFD_SUPPORT_OPENGL |             // support OpenGL 
    PFD_DOUBLEBUFFER   |             // double buffered 
    PFD_TYPE_RGBA,                   // RGBA type 
    24,                              // 24-bit color depth 
    0, 0, 0, 0, 0, 0,                // color bits ignored 
    0,                               // no alpha buffer 
    0,                               // shift bit ignored 
    0,                               // no accumulation buffer 
    0, 0, 0, 0,                      // accum bits ignored 
    32,                              // 32-bit z-buffer 
    0,                               // no stencil buffer 
    0,                               // no auxiliary buffer 
    0, //PFD_MAIN_PLANE,                  // main layer 
    0,                               // reserved 
    0, 0, 0                          // layer masks ignored 
  }; 
  int pixelFormat;

  TIMECAPS tc ;
  
  Sprite     SP ;
  GameInfo   GmInfo ;
  uint       ThreadID ;
  int        win_w ;
  int        win_h ;
  Thread     ThGameFrame ;
  
  uint  *vram ;
  int    bufw ;
  int    bufh ;

  int delegate() gamethread ;
  
  this(string name, string ClassName) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    //    winc.hInstance      = hInstance;
    winc.hIcon          = LoadIcon(null , cast(wchar*)(IDI_APPLICATION));
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszMenuName   = null;
    winc.lpszClassName  = cast(wchar*)(ClassName) ;
    super(name, ClassName) ;

    SP = new Sprite() ;
  }

  int setScreenSize(int width, int height){
    int i ;
    if( width<0 || height<0 ){
      return(-1) ;
    }
    bufw=width  ;
    bufh=height ;

    return(0) ;
  }

  int run(){
    try{
      GameLoop() ;
    }catch( Exception e ){
      writef("GameLoopError: %s\n",e.msg) ;
      SendMessage( this.hwnd, WM_CLOSE, 0, 0 ) ;      
    }
    return(0) ;
  }
  
  int GameLoop(){
    return(0) ;
  }
  
  int CloseProc(){
    return(0) ;
  }
    
  int DestroyProc(){
    return(0) ;
  }
    
  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      ThGameFrame.pause() ;
      DestroyProc() ;
      
      timeEndPeriod(tc.wPeriodMin);       
      wglDeleteContext( hRC );
      ReleaseDC(hwnd, WinDC);
      PostQuitMessage(0);
      break ;
      
    case WM_CREATE:
      RECT rc ;
      GetClientRect( hwnd, &rc ) ;
      win_w = rc.right  ;
      win_h = rc.bottom ;
      
      WinDC=GetDC(hwnd);
      pixelFormat = ChoosePixelFormat(WinDC,&pfd);
      SetPixelFormat(WinDC,pixelFormat,&pfd);
      hRC=wglCreateContext(WinDC);
      
      wglMakeCurrent( WinDC, hRC );
      glextInit() ;
      wglMakeCurrent( null, null );
      
      this.hInstance = (cast(LPCREATESTRUCT)(lp)).hInstance ;

      GmInfo.hwnd         = hwnd ;
      GmInfo.hdc          = WinDC;
      GmInfo.hRC          = hRC  ;
      GmInfo.SP           = SP   ;
      GmInfo.screenWidth  = bufw ;
      GmInfo.screenHeight = bufh ;
      GmInfo.stick        = 0    ;
      GmInfo.trig         = 0    ;
      GmInfo.vram         = vram ;

      gamethread = &run ;
      ThGameFrame = new Thread( gamethread ) ;
      ThGameFrame.start() ;

      timeGetDevCaps( &tc , TIMECAPS.sizeof ) ;
      timeBeginPeriod( tc.wPeriodMin ) ;
      break ;
      
    case WM_PAINT:
      PAINTSTRUCT ps ;

      wglMakeCurrent( WinDC, hRC );
      SP.drawSP(win_w, win_h) ;
      SwapBuffers(WinDC);
      wglMakeCurrent( null, null );
      BeginPaint(hwnd , &ps);
      EndPaint(hwnd , &ps);
      break ;

    case WM_SIZE:
      win_w=LOWORD(lp) ;
      win_h=HIWORD(lp) ;
      GmInfo.windowWidth  = win_w ;
      GmInfo.windowHeight = win_h ;
      break;

    case WM_MOUSEMOVE:
      {
        POINT point;
        GetCursorPos(&point);
        ScreenToClient( hwnd, &point );
	GmInfo.mousex = point.x ;
        GmInfo.mousey = point.y ;
      }
      break ;
      
    case WM_KEYDOWN:
      if(       wp == VK_LEFT  ){
	GmInfo.stick |= 0x01 ;
      }else if( wp == VK_RIGHT ){
	GmInfo.stick |= 0x02 ;
      }else if( wp == VK_DOWN ){
	GmInfo.stick |= 0x04 ;
      }else if( wp == VK_UP ){
	GmInfo.stick |= 0x08 ;
      }
      if( wp == VK_SPACE ){
	GmInfo.trig  |= 0x10 ;
      }
      if( wp == 0x58 ){ /* 'X' */
	GmInfo.trig  |= 0x80 ;
      }
      if( wp == 0x5a ){ /* 'Z' */
	GmInfo.trig  |= 0x40 ;
      }
      if( wp == 0x43 ){ /* 'C' */
	GmInfo.trig  |= 0x20 ;
      }
      break ;
      
    case WM_KEYUP:
      if(       wp == VK_LEFT  ){
	GmInfo.stick &= ~(0x01) ;
      }else if( wp == VK_RIGHT ){
	GmInfo.stick &= ~(0x02) ;
      }else if( wp == VK_DOWN ){
	GmInfo.stick &= ~(0x04) ;
      }else if( wp == VK_UP ){
	GmInfo.stick &= ~(0x08) ;
      }
      if( wp == VK_SPACE ){
	GmInfo.trig  &= ~(0x10) ;
      }
      if( wp == 0x58 ){ /* 'X' */
	GmInfo.trig  &= ~(0x80) ;
      }
      if( wp == 0x5a ){ /* 'Z' */
	GmInfo.trig  &= ~(0x40) ;
      }
      if( wp == 0x43 ){ /* 'C' */
	GmInfo.trig  &= ~(0x20) ;
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

}
