import std.math;
import Vec3D;
import Quaternion;

struct Block{
  int col ;
  int row ;
  int lay ;
  Quaternion rot ;
} ;

struct Anim{
  vec3D      pos ;
  Quaternion rot ;
} ;
  

class RubiksCube{

  Block[][] cube  ;
  Block[]   wklay ;
  int       num   ;

  this(){
  }
  
  ~this(){
    delete cube ;
    delete wklay ;
  }

  int init(int num){
    if( num<2 ) return(-1) ;
    delete cube ;
    delete wklay ;
    cube.length=num ;
    int pos=0 ;
    for( int l=0 ; l<cube.length ; l++ ){
      cube[l].length = num*num ;
      for( int r=0 ; r<num ; r++ ){
	for( int c=0 ; c<num ; c++ ){
	  cube[l][r*num+c].col=c ;
	  cube[l][r*num+c].row=r ;
	  cube[l][r*num+c].lay=l ;
	  cube[l][r*num+c].rot.setRotate(0.0, 1.0, 0.0, 0.0) ;
	}
      }
    }
    wklay.length=num*num ;
    this.num=num ;
    return(0) ;
  }

  Quaternion getBlockRotate(int cnum, int rnum, int lnum){
    return(cube[lnum][rnum*num+cnum].rot) ;
  }

  
  int rotate(int type, int num, int rot){
    int ret=-1 ;
    switch(type){
    case 0:
      ret=rotColumn(num,rot) ;
      break ;
    case 1:
      ret=rotRow(num,rot) ;
      break ;
    case 2:
      ret=rotLayer(num,rot) ;
      break ;
    default :
      break ;
    }
    return(ret) ;
  }

  
  int rotColumn(int cnum, int rot){
    Quaternion rot_axis ;
    if( cnum<0 || cnum>=num ) return(-1) ;

    for( int l=0, p=0 ; l<num ; l++ ){
      for( int r=0 ; r<num ; r++ ){
	wklay[p++]=cube[l][r*num+cnum] ;
      }
    }
    switch(rot%4){
    case 1 :
      for( int r=0, p=0 ; r<num ; r++ ){
	for( int l=0 ; l<num ; l++ ){
	  rot_axis.setRotate(-PI*2.0/4.0, 1.0, 0.0, 0.0) ;
	  cube[num-1-l][r*num+cnum]=wklay[p++] ;
	  rot_axis.mul(&cube[num-1-l][r*num+cnum].rot) ;
	  cube[num-1-l][r*num+cnum].rot.set(&rot_axis) ;
	}
      }
      break ;
    case 2 :
      for( int l=0, p=0 ; l<num ; l++ ){
	for( int r=0 ; r<num ; r++ ){
	  rot_axis.setRotate(-PI*2.0/2.0, 1.0, 0.0, 0.0) ;
	  cube[num-1-l][(num-1-r)*num+cnum]=wklay[p++] ;
	  rot_axis.mul(&cube[num-1-l][(num-1-r)*num+cnum].rot) ;
	  cube[num-1-l][(num-1-r)*num+cnum].rot.set(&rot_axis) ;
	}
      }
      break ;
    case 3 :
      for( int r=0, p=0 ; r<num ; r++ ){
	for( int l=0 ; l<num ; l++ ){
	  rot_axis.setRotate(-PI*2.0*3.0/4.0, 1.0, 0.0, 0.0) ;
	  cube[l][(num-1-r)*num+cnum]=wklay[p++] ;
	  rot_axis.mul(&cube[l][(num-1-r)*num+cnum].rot) ;
	  cube[l][(num-1-r)*num+cnum].rot.set(&rot_axis) ;
	}
      }
      break ;
    default :
      break ;
    }
    return(0) ;
  }
  
  int rotRow(int rnum, int rot){
    Quaternion rot_axis ;
    if( rnum<0 || rnum>=num ) return(-1) ;

    for( int l=0, p=0 ; l<num ; l++ ){
      for( int c=0 ; c<num ; c++ ){
	wklay[p++]=cube[l][rnum*num+c] ;
      }
    }
      
    switch(rot%4){
    case 1 :
      for( int c=0, p=0 ; c<num ; c++ ){
	for( int l=0 ; l<num ; l++ ){
	  rot_axis.setRotate(PI*2.0/4.0, 0.0, 1.0, 0.0) ;
	  cube[num-1-l][rnum*num+c]=wklay[p++] ;
	  rot_axis.mul(&cube[num-1-l][rnum*num+c].rot) ;
	  cube[num-1-l][rnum*num+c].rot.set(&rot_axis) ;
	}
      }
      break ;
    case 2 :
      for( int l=0, p=0 ; l<num ; l++ ){
	for( int c=0 ; c<num ; c++ ){
	  rot_axis.setRotate(PI*2.0/2.0, 0.0, 1.0, 0.0) ;
	  cube[num-1-l][rnum*num+(num-1-c)]=wklay[p++] ;
	  rot_axis.mul(&cube[num-1-l][rnum*num+(num-1-c)].rot) ;
	  cube[num-1-l][rnum*num+(num-1-c)].rot.set(&rot_axis) ;
	}
      }
      break ;
    case 3 :
      for( int c=0, p=0 ; c<num ; c++ ){
	for( int l=0 ; l<num ; l++ ){
	  rot_axis.setRotate(PI*2.0*3.0/4.0, 0.0, 1.0, 0.0) ;
	  cube[l][rnum*num+(num-1-c)]=wklay[p++] ;
	  rot_axis.mul(&cube[l][rnum*num+(num-1-c)].rot) ;
	  cube[l][rnum*num+(num-1-c)].rot.set(&rot_axis) ;
	}
      }
      break ;
    default :
      break ;
    }
    return(0) ;
  }

  int rotLayer(int lnum, int rot){
    Quaternion rot_axis ;
    if( lnum<0 || lnum>=num ) return(-1) ;

    for( int r=0, p=0 ; r<num ; r++ ){
      for( int c=0 ; c<num ; c++ ){
	wklay[p++]=cube[lnum][r*num+c] ;
      }
    }
      
    switch(rot%4){
    case 1 :
      for( int c=0, p=0 ; c<num ; c++ ){
	for( int r=0 ; r<num ; r++ ){
	  rot_axis.setRotate(-PI*2.0/4.0, 0.0, 0.0, 1.0) ;
	  cube[lnum][(num-1-r)*num+c]=wklay[p++] ;
	  rot_axis.mul(&cube[lnum][(num-1-r)*num+c].rot) ;
	  cube[lnum][(num-1-r)*num+c].rot.set(&rot_axis) ;
	}
      }
      break ;
    case 2 :
      for( int r=0, p=0 ; r<num ; r++ ){
	for( int c=0 ; c<num ; c++ ){
	  rot_axis.setRotate(-PI*2.0/2.0, 0.0, 0.0, 1.0) ;
	  cube[lnum][(num-1-r)*num+(num-1-c)]=wklay[p++] ;
	  rot_axis.mul(&cube[lnum][(num-1-r)*num+(num-1-c)].rot) ;
	  cube[lnum][(num-1-r)*num+(num-1-c)].rot.set(&rot_axis) ;
	}
      }
      break ;
    case 3 :
      for( int c=0, p=0 ; c<num ; c++ ){
	for( int r=0 ; r<num ; r++ ){
	  rot_axis.setRotate(-PI*2.0*3.0/4.0, 0.0, 0.0, 1.0) ;
	  cube[lnum][r*num+(num-1-c)]=wklay[p++] ;
	  rot_axis.mul(&cube[lnum][r*num+(num-1-c)].rot) ;
	  cube[lnum][r*num+(num-1-c)].rot.set(&rot_axis) ;
	}
      }
      break ;
    default :
      break ;
    }
    return(0) ;
  }

  int getNum(){
    return(num) ;
  }

  Anim[] makeAnimation(int type,int rnum,int rot,double blksize, double rate){
    Anim[] anim ;
    anim.length=(num*num*num) - (num-2)*(num-2)*(num-2) ;

    int ptr=0 ;
    for( int z=0 ; z<num ; z++ ){
      for( int y=0 ; y<num ; y++ ){
	for( int x=0 ; x<num ; x++ ){
	  if( x==0 || y==0 || z==0 || x==num-1 || y==num-1 || z==num-1 ){
	    anim[ptr].pos.set(cast(double)(x)*blksize-(blksize*(num-1)/2.0),
			      cast(double)(y)*blksize-(blksize*(num-1)/2.0),
			      cast(double)(z)*blksize-(blksize*(num-1)/2.0)) ;
	    Quaternion blk_rot=getBlockRotate(x,y,z) ;
	    anim[ptr].rot=blk_rot ;
	    
	    switch(type){
	    case 0:
	      if( x==rnum ){
		vec3D      vrot_axis ;
		vrot_axis.set(-1.0, 0.0, 0.0) ;
		anim[ptr].pos.rotate(PI*2.0*rate*cast(double)(rot==3?-1:rot)/4.0,vrot_axis) ;
		Quaternion qrot_axis ;
		qrot_axis.setRotate(PI*2.0*rate*cast(double)(rot==3?-1:rot)/4.0, -1.0, 0.0, 0.0) ;
		qrot_axis.mul(&blk_rot) ;
		anim[ptr].rot=qrot_axis ;
	      }
	      break ;
	    case 1:
	      if( y==rnum ){
		vec3D      vrot_axis ;
		vrot_axis.set(0.0, 1.0, 0.0) ;
		anim[ptr].pos.rotate(PI*2.0*rate*cast(double)(rot==3?-1:rot)/4.0,vrot_axis) ;
		Quaternion qrot_axis ;
		qrot_axis.setRotate(PI*2.0*rate*cast(double)(rot==3?-1:rot)/4.0, 0.0, 1.0, 0.0) ;
		qrot_axis.mul(&blk_rot) ;
		anim[ptr].rot=qrot_axis ;
	      }
	      break ;
	    case 2:
	      if( z==rnum ){
		vec3D      vrot_axis ;
		vrot_axis.set(0.0, 0.0, -1.0) ;
		anim[ptr].pos.rotate(PI*2.0*rate*cast(double)(rot==3?-1:rot)/4.0,vrot_axis) ;
		Quaternion qrot_axis ;
		qrot_axis.setRotate(PI*2.0*rate*cast(double)(rot==3?-1:rot)/4.0, 0.0, 0.0, -1.0) ;
		qrot_axis.mul(&blk_rot) ;
		anim[ptr].rot=qrot_axis ;
	      }
	      break ;
	    default :
	      break ;
	    }
	    ptr++ ;
	  }
	}
      }
    }
    return(anim) ;
  }
    
}
