import std.c.stdlib;
import std.stdio;
import std.string;
import win32.windows;
import win32frame.glut;
import util.Image;

import OBJ3D;
import GameInfo ;

//extern (C) char* load_png(char *file, int *width, int *height, int *bpp);

struct PCG2D{
  uint[] img        ;
  int    width  = 0 ;
  int    height = 0 ;
  uint   texname    ;
}

enum BLEND_TYPE{
  normal=0,
  add   =1
}
  
struct SP2D{
  int        no ;
  int        x  ;
  int        y  ;
  int        alpha ;
  double     rd ;
  double     sx, sy ;
  BLEND_TYPE blend ;
}  

struct PCG3D{
  uint  dsp ;
}

struct SP3D{
  int    no ;
  BLEND_TYPE blend ;
  double   x,  y,  z     ;
  double  rx, ry, rz, rd ;
  double  sx, sy, sz     ;
}  


class Sprite
{
  int     initialize ;
  GameInfo* GmInfo ;
  
  PCG2D[] pcg2d ;
  SP2D[]  sp2d  ;
  int     cur2dsp ;

  PCG3D[] pcg3d ;
  SP3D[]  sp3d  ;
  int     pcg3d_top ;
  int     cur3dsp ;
  
  int     tick ;
  int     starttime , nowtime, beforetime, difftime;
  int     calc_fps_begin ;
  int     fps ;
  int     baseSPF ; /** 1フレームあたりの時間 **/
  int     pause_state ;

  uint    _3d_layer ;

  double  eyex, eyey, eyez ;
  double  cntx, cnty, cntz ;
  double   upx,  upy,  upz ;
  double  fovy, zNear,zFar ;

  GLfloat[] mat_diffuse  ;
  GLfloat[] mat_ambient  ;
  GLfloat[] mat_specular ;
  GLfloat[] light0pos    ;
  GLfloat[] C_Table ;
  

  this(){
    eyex=0.0 ; eyey=0.0 ; eyez=3.73205 ;
    cntx=0.0 ; cnty=0.0 ; cntz=0.0 ;
    upx =0.0 ; upy =1.0 ; upz =0.0 ;
    fovy=30.0;
    zNear=0.3;
    zFar =200.0 ;
    synchronized{initialize=0 ;}
    pause_state=0 ;
  }

  int init(int max2dpcg, int max2dsp, int max3dpcg, int max3dsp, GameInfo* GI){
    if( initialize==1 ){
      throw new Exception("Class Sprite.init() multi calling") ;
    }
    
    fps=0 ;

    pcg2d.length=max2dpcg ;
    sp2d.length =max2dsp  ;
    cur2dsp     =0 ;
    
    pcg3d.length=max3dpcg ;
    for( int i=0 ; i<pcg3d.length ; i++ ){
      pcg3d[i].dsp=0 ;
    }
    sp3d.length =max3dsp  ;
    cur3dsp     =0 ;
    
    tick        =0 ;
    beforetime  =0 ;
    GmInfo      =GI ;

    int ret=wglMakeCurrent( GmInfo.hdc, GmInfo.hRC ) ;
    if( ret==false ){
      throw new Exception("Failed wglMakeCurrent") ;
    }
    glEnable(GL_DEPTH_TEST);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    glGenTextures(1,&_3d_layer) ;
    glBindTexture(GL_TEXTURE_2D, _3d_layer) ;
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR ) ; //GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR ) ; //GL_NEAREST);

    pcg3d_top=glGenLists(max3dpcg) ;
    if( pcg3d_top==0 ){
      writef("glGenLists faild\n") ;
    }
    
    wglMakeCurrent( null, null );

    mat_diffuse.length=8 ;
    mat_diffuse[0]=0.4 ;
    mat_diffuse[1]=0.4 ;
    mat_diffuse[2]=0.5 ;
    mat_diffuse[3]=1.0 ;
    mat_ambient.length=8 ;
    mat_ambient[0]=0.5 ;
    mat_ambient[1]=0.5 ;
    mat_ambient[2]=0.5 ;
    mat_ambient[3]=1.0 ;
    mat_specular.length=8 ;
    mat_specular[0]=0.7 ;
    mat_specular[1]=0.7 ;
    mat_specular[2]=0.7 ;
    mat_specular[3]=1.0 ;
    light0pos.length=8 ;
    light0pos[0]= 1.0 ;
    light0pos[1]= 1.0 ;
    light0pos[2]= 4.5 ;
    light0pos[3]= 1.0 ;

    C_Table.length = 32 ;
    
    synchronized{initialize=1 ;}
    baseSPF = 16 ; //16ms
    starttime = timeGetTime() ;
    
    return(0) ;
  }

  int resetSync(){
    beforetime = timeGetTime() ;
    if( tick==0 ){
      calc_fps_begin=beforetime ;
    }
    if( pause()==0 ){
      flush() ;
    }
    return(pause_state) ;
  }

  int waitSync(){
    int needframe ;

    RedrawWindow( GmInfo.hwnd, null, null, RDW_INVALIDATE|RDW_UPDATENOW ) ;
    //    SendMessage( GmInfo.hwnd, WM_PAINT, 0, 0 ) ;
    
    nowtime = timeGetTime() ;
    difftime= nowtime - beforetime ;
    //    writef("befor difftime=%d\n",difftime) ;

    //needframe=baseSPF ; // 1フレーム周期以内の場合のみ同期
    for( needframe=baseSPF ; (difftime > needframe) ; needframe+=baseSPF ){} ; //フレーム表示周期で同期化
    if( (needframe-difftime)>0 ) Sleep(needframe-difftime) ;
    
    nowtime = timeGetTime() ;
    difftime= nowtime - beforetime ;
    //    writef("after difftime=%d\n",difftime) ;
    
    //    while(-1){
    //      nowtime = timeGetTime() ;
    //      difftime= nowtime - beforetime ;
    //      if( difftime>12 ) break ;
    //    }
    
    tick = (tick + 1)%256 ;
    if( tick==255 ){
      fps = nowtime - calc_fps_begin ;
    }
    return(0) ;
  }

  void flush(){
    cur2dsp=0 ;
    cur3dsp=0 ;
  }

  int pause(){
    int keysns ;
    synchronized { keysns = (GmInfo.trig & 0x10) ; }
    if( keysns!=0 ){
      while(-1){
	synchronized { keysns = (GmInfo.trig & 0x10) ; }
	if( keysns==0 ){
	  pause_state = (pause_state + 1) %2 ;
	  break ;
	}
      }
    }
    return(pause_state) ;
  }
  
  
  int set2DSP(int pcgno, int x, int y){
    if( pcgno<0 ) return(-1) ;
    if( cur2dsp < sp2d.length ){
      int spno ;
      sp2d[cur2dsp].no=pcgno  ;
      sp2d[cur2dsp].x =x ;
      sp2d[cur2dsp].y =y ;
      sp2d[cur2dsp].alpha=255 ;
      sp2d[cur2dsp].blend=BLEND_TYPE.normal ;
      sp2d[cur2dsp].rd=0      ;
      sp2d[cur2dsp].sx=1.0    ;
      sp2d[cur2dsp].sy=1.0    ;
      
      spno=cur2dsp ;
      cur2dsp++ ;
      return(spno) ;
    }
    return(-1) ;
  }

  int set2DSP(int pcgno, int x, int y, double rd, int a, double scale){
    if( pcgno<0 ) return(-1) ;
    if( cur2dsp < sp2d.length ){
      int spno ;
      sp2d[cur2dsp].no=pcgno  ;
      sp2d[cur2dsp].x =x ;
      sp2d[cur2dsp].y =y ;
      sp2d[cur2dsp].alpha=a   ;
      sp2d[cur2dsp].blend=BLEND_TYPE.normal ;
      sp2d[cur2dsp].rd=rd     ;
      sp2d[cur2dsp].sx=scale  ;
      sp2d[cur2dsp].sy=scale  ;
      
      spno=cur2dsp ;
      cur2dsp++ ;
      return(spno) ;
    }
    return(-1) ;
  }

  int set2DSP(int pcgno, int x, int y, double rd, int a, BLEND_TYPE bltype, double scalex, double scaley){
    if( pcgno<0 ) return(-1) ;
    if( cur2dsp < sp2d.length ){
      int spno ;
      sp2d[cur2dsp].no=pcgno ;
      sp2d[cur2dsp].x =x ;
      sp2d[cur2dsp].y =y ;
      sp2d[cur2dsp].alpha=a  ;
      sp2d[cur2dsp].blend=bltype ;
      sp2d[cur2dsp].rd=rd    ;
      sp2d[cur2dsp].sx=scalex;
      sp2d[cur2dsp].sy=scaley;
      
      spno=cur2dsp ;
      cur2dsp++ ;
      return(spno) ;
    }
    return(-1) ;
  }

  int set3DSP(int pcgno, double x, double y, double z){
    if( pcgno<0 ) return(-1) ;
    if( cur3dsp < sp3d.length ){
      int spno ;
      sp3d[cur3dsp].no=pcgno ;
      sp3d[cur3dsp].x =((x)-(GmInfo.screenWidth /2)) /(GmInfo.screenHeight/2.0) ;
      sp3d[cur3dsp].y =((y)-(GmInfo.screenHeight/2))/-(GmInfo.screenHeight/2.0) ;
      sp3d[cur3dsp].z = (z)                         /-(GmInfo.screenHeight/2.0) ;
      sp3d[cur3dsp].rx=0.0   ;
      sp3d[cur3dsp].ry=0.0   ;
      sp3d[cur3dsp].rz=0.0   ;
      sp3d[cur3dsp].rd=0.0   ;
      sp3d[cur3dsp].sx=1.0   ;
      sp3d[cur3dsp].sy=1.0   ;
      sp3d[cur3dsp].sz=1.0   ;
      sp3d[cur3dsp].blend=BLEND_TYPE.normal ;
      
      spno=cur3dsp ;
      cur3dsp++ ;
      return(spno) ;
    }
    return(-1) ;
  }
  
  int set3DSP(int pcgno, double x, double y, double z, double rx, double ry, double rz, double rd,
	      double scalex, double scaley, double scalez, BLEND_TYPE bltype ){
    if( pcgno<0 ) return(-1) ;
    if( cur3dsp < sp3d.length ){
      int spno ;
      sp3d[cur3dsp].no=pcgno ;
      sp3d[cur3dsp].x =((x)-(GmInfo.screenWidth /2)) /(GmInfo.screenHeight/2.0) ;
      sp3d[cur3dsp].y =((y)-(GmInfo.screenHeight/2))/-(GmInfo.screenHeight/2.0) ;
      sp3d[cur3dsp].z = (z)                         /-(GmInfo.screenHeight/2.0) ;
      sp3d[cur3dsp].rx=rx    ;
      sp3d[cur3dsp].ry=ry    ;
      sp3d[cur3dsp].rz=rz    ;
      sp3d[cur3dsp].rd=rd    ;
      sp3d[cur3dsp].sx=scalex;
      sp3d[cur3dsp].sy=scaley;
      sp3d[cur3dsp].sz=scalez;
      sp3d[cur3dsp].blend=bltype ;
      
      spno=cur3dsp ;
      cur3dsp++ ;
      return(spno) ;
    }
    return(-1) ;
  }

  
  int set3Draw(int pcgno, double x, double y, double z){
    if( pcgno<0 ) return(-1) ;
    if( cur3dsp < sp3d.length ){
      int spno ;
      sp3d[cur3dsp].no=pcgno ;
      sp3d[cur3dsp].x =x     ;
      sp3d[cur3dsp].y =y     ;
      sp3d[cur3dsp].z =z     ;
      sp3d[cur3dsp].rx=0.0   ;
      sp3d[cur3dsp].ry=0.0   ;
      sp3d[cur3dsp].rz=0.0   ;
      sp3d[cur3dsp].rd=0.0   ;
      sp3d[cur3dsp].sx=1.0   ;
      sp3d[cur3dsp].sy=1.0   ;
      sp3d[cur3dsp].sz=1.0   ;
      sp3d[cur3dsp].blend=BLEND_TYPE.normal ;
      
      spno=cur3dsp ;
      cur3dsp++ ;
      return(spno) ;
    }
    return(-1) ;
  }

  
  int set3Draw(int pcgno, double x, double y, double z, double rx, double ry, double rz, double rd,
	       double scalex, double scaley, double scalez, BLEND_TYPE bltype ){
    if( pcgno<0 ) return(-1) ;
    if( cur3dsp < sp3d.length ){
      int spno ;
      sp3d[cur3dsp].no=pcgno ;
      sp3d[cur3dsp].x =x     ;
      sp3d[cur3dsp].y =y     ;
      sp3d[cur3dsp].z =z     ;
      sp3d[cur3dsp].rx=rx    ;
      sp3d[cur3dsp].ry=ry    ;
      sp3d[cur3dsp].rz=rz    ;
      sp3d[cur3dsp].rd=rd    ;
      sp3d[cur3dsp].sx=scalex;
      sp3d[cur3dsp].sy=scaley;
      sp3d[cur3dsp].sz=scalez;
      sp3d[cur3dsp].blend=bltype ;
      
      spno=cur3dsp ;
      cur3dsp++ ;
      return(spno) ;
    }
    return(-1) ;
  }

  
  int add2DPCG(string fname, int pcgno){
    int    i ;
    uint[] img ;
    int    imgw,imgh ;

    if( pcgno < 0 || pcgno>=pcg2d.length ){
      return(-1) ;
    }

    Image image = new Image() ;
    if( image.open(fname)<0 ){
      writef("can not load texture png (%s)\n",fname) ;
      image.close() ;
      delete image ;
      return(-1) ;
    }
    imgw = image.getWidth() ;
    imgh = image.getHeight() ;
    img  = image.readline(imgh) ;

    pcg2d[pcgno].img    = img ;
    pcg2d[pcgno].width  = imgw ;
    pcg2d[pcgno].height = imgh ;
    image.close() ;
    delete image ;
    
    wglMakeCurrent( GmInfo.hdc, GmInfo.hRC ) ;
    glGenTextures(1,&pcg2d[pcgno].texname) ;
    glBindTexture(GL_TEXTURE_2D,pcg2d[pcgno].texname) ;
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE) ; //GL_REPEAT) ;
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE) ; //GL_REPEAT) ;
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST) ;
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST) ; //GL_NEAREST) ;
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL) ;
    //    glTexImage2D(GL_TEXTURE_2D,0, GL_RGBA,
    //		 pcg2d[pcgno].width,
    //		 pcg2d[pcgno].height,
    //		 0, GL_RGBA, GL_UNSIGNED_BYTE,
    //		 cast(char*)(pcg2d[pcgno].img)) ;
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA,
                      pcg2d[pcgno].width, pcg2d[pcgno].height,
                      GL_RGBA, GL_UNSIGNED_BYTE,
                      cast(char*)(pcg2d[pcgno].img)) ;   
    
    wglMakeCurrent( null, null );
    
    return(0) ;
  }

  int add2DPCG(string fname,int pcgno, int w, int h, int xnum, int ynum){
    int    i,j,k ;
    int    x,y ;
    int    curpcg ;
    uint[] img ;
    int    imgw,imgh,bpp ;
    int    clipxnum, clipynum ;

    if( pcgno < 0 || pcgno>=pcg2d.length ){
      return(-1) ;
    }
    
    Image image = new Image() ;
    if( image.open(fname)<0 ){
      writef("can not load texture png (%s)\n",fname) ;
      image.close() ;
      delete image ;
      return(-1) ;
    }
    imgw = image.getWidth() ;
    imgh = image.getHeight() ;
    img  = image.readline(imgh) ;
    
    clipxnum = (imgw/w)*imgw ; clipxnum = (clipxnum<xnum) ? clipxnum : xnum ;
    clipynum = (imgh/h)*imgh ; clipynum = (clipynum<ynum) ? clipynum : ynum ;
    
    curpcg=pcgno ;
    
    for( i=0 ; i<clipynum ; i++ ){
      for( j=0 ; j<clipxnum ; j++ ){
	pcg2d[curpcg].img.length = w*h ;
	pcg2d[curpcg].width      = w   ;
	pcg2d[curpcg].height     = h   ;
	for( y=0 ; y<h ; y++ ){
	  pcg2d[curpcg].img[(y*w)..(y*w+w)]
	    = img[(((y+(i*h))*imgw)+j*w)..(((y+(i*h))*imgw)+j*w+w)] ;
	}
	wglMakeCurrent( GmInfo.hdc, GmInfo.hRC ) ;
	glGenTextures(1,&pcg2d[curpcg].texname) ;
	glBindTexture(GL_TEXTURE_2D,pcg2d[curpcg].texname) ;
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE) ; //GL_REPEAT) ;
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE) ; //GL_REPEAT) ;
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST) ;
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST) ; //GL_NEAREST) ;
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL) ;
	//	glTexImage2D(GL_TEXTURE_2D,0, GL_RGBA,
	//		     pcg2d[curpcg].width,
	//		     pcg2d[curpcg].height,
	//		     0, GL_RGBA, GL_UNSIGNED_BYTE,
	//		     cast(char*)(pcg2d[curpcg].img)) ;
	gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA,
			  pcg2d[curpcg].width, pcg2d[curpcg].height,
			  GL_RGBA, GL_UNSIGNED_BYTE,
			  cast(char*)(pcg2d[curpcg].img)) ;   
	wglMakeCurrent( null, null );
	
	curpcg++ ;
      }
    }
    delete img ;
    image.close() ;
    delete image ;
	
    return(0) ;
  }

  int add3DPCG(string fname, int pcgno){
    int i ;
    OBJ3D  obj3d ;

    if( pcgno<0 || pcgno>=pcg3d.length ){
      return(-1) ;
    }
    
    obj3d = new OBJ3D() ;
    
    obj3d.open(fname) ;
    obj3d.read() ;

    wglMakeCurrent( GmInfo.hdc, GmInfo.hRC ) ;
    glLoadIdentity();

    pcg3d[pcgno].dsp=pcg3d_top+pcgno ;


    foreach( mtlid ; obj3d.obj.matNameToID ){
      if( obj3d.obj.mat[mtlid].tex.length!=0 ){
	uint[] img ;
	int    imgw,imgh ;
	
	Image image = new Image() ;
	if( image.open(obj3d.obj.mat[mtlid].tex)<0 ){
	  writef("can not load texture png (%s)\n",fname) ;
	  image.close() ;
	  delete image ;
	  return(-1) ;
	}
	imgw = image.getWidth() ;
	imgh = image.getHeight() ;
	img  = image.readline(imgh) ;
	
	obj3d.obj.mat[mtlid].teximg = img ;
	image.close() ;
	delete image ;
	
	glGenTextures(1,&obj3d.obj.mat[mtlid].texid) ;
	glBindTexture(GL_TEXTURE_2D,obj3d.obj.mat[mtlid].texid) ;
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT) ;
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT) ;
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST) ;
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST) ;
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL) ;
	glTexImage2D(GL_TEXTURE_2D,0, GL_RGBA,
		     imgw,
		     imgh,
		     0, GL_RGBA, GL_UNSIGNED_BYTE,
		     cast(char*)(obj3d.obj.mat[mtlid].teximg)) ;
	
      }
    }

    if( pcg3d[pcgno].dsp != 0 ){
      glDeleteLists(pcg3d[pcgno].dsp , 1);
      glNewList(pcg3d[pcgno].dsp, GL_COMPILE) ;
      
      foreach( mtlid ; obj3d.obj.matNameToID ){
	C_Table[0] = obj3d.obj.mat[mtlid].Kd.r ;
	C_Table[1] = obj3d.obj.mat[mtlid].Kd.g ;
	C_Table[2] = obj3d.obj.mat[mtlid].Kd.b ;
	C_Table[3] = 1.0 ;
	C_Table[4] = obj3d.obj.mat[mtlid].Ka.r ;
	C_Table[5] = obj3d.obj.mat[mtlid].Ka.g ;
	C_Table[6] = obj3d.obj.mat[mtlid].Ka.b ;
	C_Table[7] = 1.0 ;
	C_Table[8] = obj3d.obj.mat[mtlid].Ks.r ;
	C_Table[9] = obj3d.obj.mat[mtlid].Ks.g ;
	C_Table[10]= obj3d.obj.mat[mtlid].Ks.b ;
	C_Table[11]= 1.0 ;
	glMaterialfv(GL_FRONT, GL_DIFFUSE, &C_Table[0]) ;
	//glMaterialfv(GL_FRONT, GL_AMBIENT, &C_Table[4]) ;
	//glMaterialfv(GL_FRONT, GL_SPECULAR,&C_Table[8]) ;
	//glMaterialf( GL_FRONT, GL_SHININESS, 100      ) ;
	
	if( obj3d.obj.mat[mtlid].tex.length!=0 ){
	  glEnable(GL_TEXTURE_2D) ;
	  glBindTexture( GL_TEXTURE_2D, obj3d.obj.mat[mtlid].texid ) ;
	  for( i=0 ; i<obj3d.obj.faceV[mtlid].length ; i++ ){
	    glBegin( GL_TRIANGLES ) ; //GL_LINE_LOOP ) ;
	    glNormal3d(obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p0].x,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p0].y,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p0].z) ;
	    glTexCoord2d(obj3d.obj.uvm[obj3d.obj.faceT[mtlid][i].p0].u,
			 obj3d.obj.uvm[obj3d.obj.faceT[mtlid][i].p0].v) ;
	    glVertex3d(obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p0].x,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p0].y,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p0].z) ;
	    
	    glNormal3d(obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p1].x,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p1].y,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p1].z) ;
	    glTexCoord2d(obj3d.obj.uvm[obj3d.obj.faceT[mtlid][i].p1].u,
			 obj3d.obj.uvm[obj3d.obj.faceT[mtlid][i].p1].v) ;
	    glVertex3d(obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p1].x,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p1].y,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p1].z) ;
	    
	    glNormal3d(obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p2].x,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p2].y,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p2].z) ;
	    glTexCoord2d(obj3d.obj.uvm[obj3d.obj.faceT[mtlid][i].p2].u,
			 obj3d.obj.uvm[obj3d.obj.faceT[mtlid][i].p2].v) ;
	    glVertex3d(obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p2].x,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p2].y,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p2].z) ;
	    glEnd() ;
	  }
	  glDisable(GL_TEXTURE_2D) ;
	}else{
	  for( i=0 ; i<obj3d.obj.faceV[mtlid].length ; i++ ){
	    glBegin( GL_TRIANGLES ) ; //GL_LINE_LOOP ) ;
	    glNormal3d(obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p0].x,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p0].y,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p0].z) ;
	    glVertex3d(obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p0].x,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p0].y,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p0].z) ;
	    
	    glNormal3d(obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p1].x,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p1].y,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p1].z) ;
	    glVertex3d(obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p1].x,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p1].y,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p1].z) ;
	    
	    glNormal3d(obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p2].x,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p2].y,
		       obj3d.obj.nml[obj3d.obj.faceN[mtlid][i].p2].z) ;
	    glVertex3d(obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p2].x,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p2].y,
		       obj3d.obj.vtx[obj3d.obj.faceV[mtlid][i].p2].z) ;
	    glEnd() ;
	  }
	}
      }
      glEndList() ;
    }else{
      writef("failed glGenLists\n") ;
    }
    wglMakeCurrent( null, null );

    obj3d.close() ;
    delete obj3d ;
    
    return(0) ;
  }
    
  int drawSP(int win_w, int win_h){
    int     i ;
    double  zoom=cast(double)(GmInfo.screenHeight) ;

    if( initialize==0 ){
      return(0) ;
    }

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    
    glViewport(0, 0, win_w, win_h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    //      gluPerspective(90.0,  cast(double)(win_w)/cast(double)(win_h) , 0.3, 200.0);
    //      gluLookAt(0.0, 0.0, 1.0,
    //		0.0, 0.0, 0.0,
    //		0.0, 1.0, 0.0) ;
    
    gluPerspective(fovy,  cast(double)(win_w)/cast(double)(win_h) , zNear, zFar);
    gluLookAt(eyex,eyey,eyez,
	      cntx,cnty,cntz,
	      upx, upy, upz);
    
    glMatrixMode(GL_MODELVIEW);
    
    glClearColor( 0.3, 0.3, 0.3, 0.0 ) ;
    
    glDepthMask(GL_TRUE) ;
    glDisable(GL_BLEND) ;
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_POSITION, &light0pos[0]);
    glLightfv(   GL_LIGHT0,  GL_AMBIENT, &mat_ambient[0]  ) ;
    glEnable(GL_LIGHTING);

    for( i=0 ; i<cur3dsp ; i++ ){
      if( sp3d[i].blend==BLEND_TYPE.normal ){
      	glDisable(GL_BLEND) ;
      }else{
      	glEnable(GL_BLEND) ;
      	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
      }
      glPushMatrix();
      glTranslated( sp3d[i].x, sp3d[i].y, sp3d[i].z );
      glRotated( sp3d[i].rd, sp3d[i].rx, sp3d[i].ry, sp3d[i].rz ) ;
      glScaled(sp3d[i].sx, sp3d[i].sy, sp3d[i].sz );
      glCallList(pcg3d[sp3d[i].no].dsp) ;
      glPopMatrix();
    }
    glDisable(GL_BLEND) ;
    
    glBindTexture(GL_TEXTURE_2D,_3d_layer) ;
    glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,0,0,GmInfo.windowWidth, GmInfo.windowHeight,0) ;

    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);
    
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glLoadIdentity();
    
    glEnable(GL_BLEND) ;
    glDisable(GL_DEPTH_TEST);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_FALSE) ;
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); //GL_REPLACE);
    
    // BG texutre
    glEnable(GL_TEXTURE_2D) ;
    glBindTexture( GL_TEXTURE_2D, _3d_layer ) ;
    glBegin( GL_QUADS );
    glNormal3d(0.0,0.0,1.0) ;
    glTexCoord2d( 0.0, 0.0); glVertex3d(-1*(cast(double)(GmInfo.windowWidth)/cast(double)(GmInfo.windowHeight)),-1, 0.0) ;
    glTexCoord2d( 0.0, 1.0); glVertex3d(-1*(cast(double)(GmInfo.windowWidth)/cast(double)(GmInfo.windowHeight)), 1, 0.0) ;
    glTexCoord2d( 1.0, 1.0); glVertex3d( 1*(cast(double)(GmInfo.windowWidth)/cast(double)(GmInfo.windowHeight)), 1, 0.0) ;
    glTexCoord2d( 1.0, 0.0); glVertex3d( 1*(cast(double)(GmInfo.windowWidth)/cast(double)(GmInfo.windowHeight)),-1, 0.0) ;
    glEnd();
    glDisable(GL_TEXTURE_2D) ;

    for( i=cur2dsp-1 ; i>=0 ; i-- ){
      double x1,y1,x2,y2 ;
    
      x1 = cast(double)(-pcg2d[sp2d[i].no].width ) * sp2d[i].sx/(zoom) ;
      y1 = cast(double)( pcg2d[sp2d[i].no].height) * sp2d[i].sy/(zoom) ;
      x2 = cast(double)( pcg2d[sp2d[i].no].width ) * sp2d[i].sx/(zoom) ;
      y2 = cast(double)(-pcg2d[sp2d[i].no].height) * sp2d[i].sy/(zoom) ;

      if( sp2d[i].blend==BLEND_TYPE.normal ){
      	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      }else{
      	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
      }
      glEnable(GL_TEXTURE_2D) ;
      glBindTexture( GL_TEXTURE_2D, pcg2d[sp2d[i].no].texname ) ;
      glColor4d(1.0,1.0,1.0,cast(double)(sp2d[i].alpha)/255.0) ;
      
      glPushMatrix();
      glTranslated( cast(double)(-GmInfo.screenWidth +sp2d[i].x*2.0)/(zoom),
		    cast(double)(+GmInfo.screenHeight-sp2d[i].y*2.0)/(zoom),
		    0.0 );
      glRotated( cast(double)(sp2d[i].rd), 0.0, 0.0, 1.0 ) ;
      
      glBegin( GL_POLYGON ) ;
      glNormal3d(0.0,0.0,1.0) ;
      glTexCoord2d(0.0,1.0) ; glVertex3d(x1,y2,0.0) ;
      glTexCoord2d(1.0,1.0) ; glVertex3d(x2,y2,0.0) ;
      glTexCoord2d(1.0,0.0) ; glVertex3d(x2,y1,0.0) ;
      glTexCoord2d(0.0,0.0) ; glVertex3d(x1,y1,0.0) ;
      glEnd() ;
      
      glPopMatrix();
      glDisable(GL_TEXTURE_2D) ;
    }
    glDisable(GL_BLEND) ;
    
    return(0) ;
  }
    
  int get2DNum(){
    return(cur2dsp) ;
  }
  
  int get3DNum(){
    return(cur3dsp) ;
  }

  int getFPS(){
    if( fps<=0 ){
      return(0) ;
    }
    return(256*1000/fps) ;
  }

  int getTotalTime(){
    return( nowtime - starttime ) ;
  }
}
