import win32.windows;
import std.string;
import std.stdio;
import win32frame.writek;
import win32frame.WindowObject ;

ushort LOWORD(uint l){ return cast(ushort)(l & 0xffff); }
ushort HIWORD(uint l){ return cast(ushort)(l >>> 16)  ; }

const LBUTTONDOWN   = (0x0001) ;
const LBUTTONUP     = (0x0002) ;
const LBUTTONDBLCLK = (0x0004) ;
const RBUTTONDOWN   = (0x0001) ;
const RBUTTONUP     = (0x0002) ;
const RBUTTONDBLCLK = (0x0004) ;
const ALT_KEY       = (0x0001) ;
const CNTL_KEY      = (0x0002) ;
const SHIFT_KEY     = (0x0004) ;

class Split : WindowObject {
  WindowObject[] fitWOBJ_List ;
  
  int maxobjnum    ;
  int split_mode   ;
  int sash_width=5 ;
  int mouseX    =0 ;
  int mouseY    =0 ;
  int mouseB    =0 ;
  int split_pos    ;
  int curpos       ;
  int cappos       ;
  int touching     ;
  int minpos       ;
  int maxpos       ;

  this(string name, int mode=0) {
    winc.style          = 0; //CS_HREDRAW | CS_VREDRAW ;
    this.setStyle(WS_CHILD | WS_TABSTOP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_3DFACE)) ; //cast(HBRUSH)(GetStockObject(BLACK_BRUSH));
    winc.lpszClassName  = cast(wchar*)("Split") ;
    super(name) ;
    this.registwinc() ;
    maxobjnum =2 ;
    split_mode=mode ;
    touching=0    ;
    split_pos=-1  ;
    minpos=0     ;
    maxpos=65535 ;
  }

  int addObject(WindowObject WObj, HMENU hmenu){
    HWND tmphwnd ;
    int  ret=-1 ;
    if( fitWOBJ_List.length<=maxobjnum ){
      ret = WObj.create(this.hwnd, hmenu) ;
      tmphwnd = WObj.getHWND() ;
      if( tmphwnd != null ){
	writek("add split Object : hwnd=%08x name=%s\n", cast(int)(tmphwnd), WObj.name) ;
	fitWOBJ_List.length = fitWOBJ_List.length + 1 ;
	fitWOBJ_List[(fitWOBJ_List.length-1)] = WObj ;
	SendMessage( hwnd, WM_SIZE, 0, (winH<<16 | winW) );
      }
    }
    return( ret ) ;
  }

  int setSashPos(int pos, int min_pos, int max_pos){
    if( min_pos<=max_pos && min_pos<=pos || max_pos>=pos){
      if( min_pos>=0 ){
	minpos=min_pos ;
	maxpos=max_pos ;
	split_pos=pos ;
      }
    }
    return(0) ;
  }
    
  override LRESULT preproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch( msg ){
    case WM_CREATE :
      if( split_pos<0 ){
	split_pos = split_mode==0 ? winW/2 : winH/2 ;
      }
      curpos = split_pos ;
      break ;
      
    case WM_MOVE :
      winX=LOWORD(lp) ;
      winY=HIWORD(lp) ;
      break ;

    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_MOUSEMOVE:
      {
	POINT point;

	GetCursorPos(&point);
	ScreenToClient( hwnd, &point );
	mouseX = point.x ;
	mouseY = point.y ;
	mouseX = mouseX<0 ? 0 : mouseX>=winW ? winW-1 : mouseX ;
	mouseY = mouseY<0 ? 0 : mouseY>=winH ? winH-1 : mouseY ;
	if( msg==WM_LBUTTONDOWN || msg==WM_MOUSEMOVE ){ //マウスカーソル形状を変更する
	  if( minpos!=split_pos || maxpos!=split_pos ){
	    if( split_mode==0 ){
	      SetCursor( LoadCursor(null, cast(wchar*)(IDC_SIZEWE) ) ) ;
	    }else{
	      SetCursor( LoadCursor(null, cast(wchar*)(IDC_SIZENS) ) ) ;
	    }
	  }
	}
	if( msg==WM_LBUTTONDOWN ){
	  SetCapture( hwnd ) ;
	  touching=1 ;
	  cappos = split_mode==0 ? mouseX : mouseY ;
	}
	if( msg==WM_MOUSEMOVE ){
	  if( touching!=0 ){
	    curpos = split_pos + (split_mode==0 ? mouseX : mouseY ) - cappos ;
	    curpos = curpos < minpos ? minpos : curpos>maxpos ? maxpos : curpos ;
	    if( split_mode==0 ){
	      if( (winW - sash_width)<curpos ) curpos=winW - sash_width ;
	    }else{
	      if( (winH - sash_width)<curpos ) curpos=winH - sash_width ;
	    }
	    SendMessage( hwnd, WM_SIZE, wp, (winH<<16 | winW) );
	  }
	}
	if( msg==WM_LBUTTONUP   ){
	  ReleaseCapture() ;
	  split_pos = curpos ;
	  touching=0 ;
	}
      }
      break ;
      
    case WM_SIZE :
      winW=LOWORD(lp) ;
      winH=HIWORD(lp) ;
      
      RECT rc ;
      GetClientRect( hwnd, &rc ) ;
      {
	if( split_mode==0 ){
	  int primary_window_width = (curpos) ;
	  int primary_window_height= (winH) ;
	  int pos_x = 0 ;
	  int pos_y = 0 ;
	
	  for( int i=0 ; i<fitWOBJ_List.length ; i++ ){
	    if( fitWOBJ_List[i] !is null ){
	      if( i==0 ){
		MoveWindow( cast(HWND)(fitWOBJ_List[i].getHWND()), pos_x, pos_y,
			    primary_window_width, primary_window_height, true ) ;
	      }else{
		MoveWindow( cast(HWND)(fitWOBJ_List[i].getHWND()), pos_x, pos_y,
			    (winW - sash_width ) - primary_window_width, primary_window_height, true ) ;
	      }
	      pos_x += primary_window_width + sash_width ;
	    }
	  }
	}else{
	  int primary_window_width = (winW) ;
	  int primary_window_height= (curpos) ;
	  int pos_x = 0 ;
	  int pos_y = 0 ;
	
	  for( int i=0 ; i<fitWOBJ_List.length ; i++ ){
	    if( fitWOBJ_List[i] !is null ){
	      if( i==0 ){
		MoveWindow( cast(HWND)(fitWOBJ_List[i].getHWND()), pos_x, pos_y,
			    primary_window_width, primary_window_height, true ) ;
	      }else{
		MoveWindow( cast(HWND)(fitWOBJ_List[i].getHWND()), pos_x, pos_y,
			    primary_window_width, (winH - sash_width) - primary_window_height, true ) ;
	      }
	      pos_y += primary_window_height + sash_width ;
	    }
	  }
	}
      }
      break ;
      
    case WM_COMMAND:
      SendMessage( this.getParentHWND, WM_COMMAND, wp,lp) ;
      break ;
      
    default : break ;
    }
    
    return( this.proc(hwnd , msg , wp , lp) ) ;
  }

}
